
-- this is a tool for practice and better understanding the game

require('util')

-- goals:
-- [x] show spawnpoint progress
-- [x] show aim mechanics
-- [x] show player hitbox
-- [x] pause menu (text) for configuring the trainer
-- [ ] make config transfer across savestates
-- [x] show enemy and bullet hitboxes
-- [ ] show player grenade target/hitbox
-- [ ] show enemy grenade target
-- [ ] understand fortress spawn timer logic (when does it stop?)
-- [ ] understand enemy spawn/movement/firing logic, including difficulty level
-- [ ] allow level switch (change level number, respawn)
-- [ ] allow level checkpoint set (respawn)
-- [ ] show hidden items (ladders, vest, radio, etc?)

function main()
   gui.register(draw)
   
   while true do
      emu.frameadvance()
      updatePauseMenu()
      checkCheats()
   end
end

-- https://nesdev-wiki.nes.science/wikipages/PPU_registers.xhtml
-- https://everything2.com/title/NES+PPU
--memory.register(0x4014, 1, oamDMA)
function oamDMA(address, size, value)
   emu.message('dma '..tostring(value))
end

bulletHitbox = {x=0, y=0, w=0, h=0}
playerHitbox = {x=5, y=5, w=7, h=7}
guyHitbox = {x=0, y=0, w=16, h=15}
rocketmanHitbox = {x=0, y=8, w=16, h=15}

hitboxData = {
   -- [tile index] = hitbox,
   [0x16] = bulletHitbox, -- bullet
   [0x18] = bulletHitbox, -- big bullet
   
   [0x20] = guyHitbox, -- guy moving right
   [0x22..'f'] = guyHitbox, -- guy moving left
   [0x24] = guyHitbox, -- guy moving upright
   [0x26..'f'] = guyHitbox, -- guy moving upleft
   [0x28] = guyHitbox, -- guy moving up
   [0x2c] = guyHitbox, -- guy moving downright
   [0x2e..'f'] = guyHitbox, -- guy moving downleft
   [0x30] = guyHitbox, -- guy moving down

   [0xc0] = guyHitbox, -- knifeguy moving right
   [0xc2..'f'] = guyHitbox, -- knifeguy moving left
   [0xc4] = guyHitbox, -- knifeguy moving upright
   [0xc6..'f'] = guyHitbox, -- knifeguy moving upleft
   [0xc8] = guyHitbox, -- knifeguy moving up
   [0xcc] = guyHitbox, -- knifeguy moving downright
   [0xce..'f'] = guyHitbox, -- knifeguy moving downleft
   [0xd0] = guyHitbox, -- knifeguy moving down

   -- d0 is also knifeguy moving down. see 'seenRocketman'
   --[0xd0] = rocketmanHitbox,
   [0xd8] = rocketmanHitbox,
   [0xf0] = rocketmanHitbox,
   [0xf8] = bulletHitbox, -- rocket
}

-- a frame timer for when the last definite rocketman was seen
seenRocketman = 0
function isRocketmanTile(tile)
   return tile == 0xd8 or tile == 0xf0
end

function draw()
   local y = 8
   
   if infoOptions.spawnPointInfo then
      gui.text(8, y, string.format('sp %s', getCheckpointDistance()))
      y = y + 8
   end

   local pos = getPlayerPos()
   if infoOptions.hitboxes then
      -- this seems to be the player's push-box
      --gui.drawbox(pos.x - 4, pos.y, pos.x + 4, pos.y + 8)

      -- the game fills sprite memory from 0x700--0x7ff
      local sawRocketman = false
      for index = 0, 0x3f do
         local tile = memory.readbyte(0x700 + index * 4 + 1)
         local attr = memory.readbyte(0x700 + index * 4 + 2)
         local flip = (bit.band(attr, 0x40) ~= 0) -- horizontal

         local box
         if not flip and hitboxData[tile] then
            box = hitboxData[tile]
         elseif flip and hitboxData[tile..'f'] then
            box = hitboxData[tile..'f']
         end

         -- player and enemies look the same, but the player is always at the beginning of the sprite list
         if index < 4 and box == guyHitbox then
            box = playerHitbox
         end
         
         -- d0 is both knifeguy and rocketman, so we need context
         if tile == 0xd0 and seenRocketman <= 4 then
            box = rocketmanHitbox
         end
         if isRocketmanTile(tile) then
            sawRocketman = true
         end

         if box then
            -- add one because "sprite data is delayed by one scanline"
            local by = memory.readbyte(0x700 + index * 4) + 1
            local bx = memory.readbyte(0x700 + index * 4 + 3)
            if box.w > 0 or box.h > 0 then
               gui.drawbox(bx + box.x, by + box.y, bx + box.x + box.w, by + box.y + box.h, nil, box.color)
            else
               gui.drawline(bx - 2, by, bx + 2, by, nil, box.color)
               gui.drawline(bx, by - 2, bx, by + 2, nil, box.color)
            end
         end
      end
      
      if sawRocketman then
         seenRocketman = 0
      elseif seenRocketman < 60 then
         seenRocketman = seenRocketman + 1
      end
   end
   
   local aim = getPlayerAim()
   local target = {x=pos.x + 16*aim.x, y=pos.y+4 + 16*aim.y}
   if infoOptions.aimGuide then
      --gui.drawbox(target.x - 4, target.y - 4, target.x + 4, target.y + 4)
      gui.drawline(pos.x, pos.y + 4, target.x, target.y)
   end

   if getGameMode() == 'pause' then
      if not pauseMenu then
         pauseMenuItems.selection = 1
         pauseMenu = {
            stack = {pauseMenuItems},
         }
      end
      
      gui.text(112, 60, 'PAUSED')

      local top = pauseMenu.stack[#pauseMenu.stack]
      for i, item in ipairs(top) do
         if i == top.selection then
            gui.text(64, 68 + i*8, '>')
         end
         local text = tostring(getOrRun(item.text))
         if item.select then
            text = text..' >'
         end
         gui.text(80, 68 + i*8, text)
      end
      
   else
      pauseMenu = nil
   end
end

function getOrRun(obj)
   if type(obj) == 'function' then
      return obj()
   else
      return obj
   end
end

pauseMenuItems = {
   {
      text = 'information',
      select = function()
         return infoOptions
      end,
   },
   {
      text = 'cheats',
      select = function()
         return cheatOptions
      end,
   },

   --[[{
      text = 'respawn',
      right = function()

         -- code e61f
         for i = 0, 3 do
            memory.writebyte(0x700 + i*4, memory.readbyte(0x84e7 + i*4) + memory.readbyte(0x64))
            memory.writebyte(0x703 + i*4, memory.readbyte(0x84e8 + i*4) + memory.readbyte(0x62))
            memory.writebyte(0x702 + i*4, memory.readbyte(0x84e9 + i*4) + memory.readbyte(0x69))
            memory.writebyte(0x701 + i*4, memory.readbyte(0x84ea + i*4))
         end

         memory.writebyte(0x7d, 0x06)
         memory.writebyte(0x7e, 0xa8)
         memory.writebyte(0x7f, 0xb2)

         for i = 0x730, 0x7c4, 4 do
            memory.writebyte(i, 0xf0)
         end
      end,
   },--]]
   
   {
      text = function()
         return 'level '..memory.readbyte(0x049d)
      end,
      left = function()
         local level = memory.readbyte(0x049d) - 1
         if level >= 0 then
            memory.writebyte(0x049d, level)
            memory.writebyte(0x049e, level % 4)
         end
      end,
      right = function()
         local level = memory.readbyte(0x049d) + 1
         if level <= 15 then
            memory.writebyte(0x049d, level)
            memory.writebyte(0x049e, level % 4)
         end
      end,
   },

   -- FIXME this corrupts the level, and doesn't stick when you die
   -- possibly related to scrolling:
   -- a1 a2 a3 aa d1 d4 d5 03xx
   {
      text = function()
         local section = 0xfd - memory.readbyte(0x7b)
         return 'section '..section
      end,
      left = function()
         local section = 0xfd - memory.readbyte(0x7b)
         if section > 0 then
            memory.writebyte(0x7b, 0xfd - (section - 1))
         end
      end,
      right = function()
         local section = 0xfd - memory.readbyte(0x7b)
         if section < 11 then
            memory.writebyte(0x7b, 0xfd - (section + 1))
         end
      end,
   },
}

function booleanOption(desc, obj, property, changeFunc)
   return {
      text = function()
         return desc..' ['..(obj[property] and 'on' or 'off')..']'
      end,
      left = function()
         obj[property] = not obj[property]
         if changeFunc then changeFunc(obj[property]) end
      end,
      right = function()
         obj[property] = not obj[property]
         if changeFunc then changeFunc(obj[property]) end
      end,
   }
end

infoOptions = {}
table.append(infoOptions, {
   booleanOption('hitboxes', infoOptions, 'hitboxes'),
   booleanOption('aim guide', infoOptions, 'aimGuide'),
   booleanOption('spawn point info', infoOptions, 'spawnPointInfo'),
})

cheatOptions = {}
table.append(cheatOptions, {
   booleanOption('infinite lives', cheatOptions, 'infiniteLives'),
   booleanOption('infinite grenades', cheatOptions, 'infiniteGrenades'),
   booleanOption('infinite vest', cheatOptions, 'infiniteVest',
      function(value)
         if not value then
            memory.writebyte(0x04ab, bit.band(memory.readbyte(0x04ab), bit.bnot(0x20)))
         end
      end
   ),
})

function updatePauseMenu()
   if not pauseMenu then
      return
   end

   local top = pauseMenu.stack[#pauseMenu.stack]
   local item = top[top.selection]
   local press = getPausePresses()
   if press.up and top.selection > 1 then
      top.selection = top.selection - 1
   end
   if press.down and top.selection < #top then
      top.selection = top.selection + 1
   end
   if press.left and item.left then
      item.left()
   end
   if press.right and item.right then
      item.right()
   end
   if press.A and item.select then
      table.insert(pauseMenu.stack, getOrRun(item.select))
      pauseMenu.stack[#pauseMenu.stack].selection = 1
   end
   if press.B and #pauseMenu.stack > 1 then
      table.remove(pauseMenu.stack)
   end
end

pauseState = joypad.get(1)
function getPausePresses()
   local pad = joypad.get(1)
   local changes = {}
   for key in pairs(pad) do
      if pad[key] ~= pauseState[key] then
         changes[key] = pad[key]
      end
   end
   pauseState = pad
   return changes
end

function checkCheats()
   if cheatOptions.infiniteLives and memory.readbyte(0x49c) ~= 3 then
      memory.writebyte(0x49c, 3)
   end

   if cheatOptions.infiniteGrenades and memory.readbyte(0x4a6) ~= 15 then
      memory.writebyte(0x4a6, 0x15)
   end
   
   if cheatOptions.infiniteVest then
      if bit.band(memory.readbyte(0x04ab), 0x20) == 0 then
         memory.writebyte(0x04ab, bit.bor(memory.readbyte(0x04ab), 0x20))
      end
      if memory.readbyte(0x95) ~= 0 then
         memory.writebyte(0x95, 0)
      end
   end
end

function getGameMode()
   return ({[0]='pause', 'unknown', 'heli', 'play', 'end'})[memory.readbyte(0xf2)]
   -- addrs F1 F2
   -- paused/menu/dead: 64 0
   -- spawning/heli: 128 2
   -- playing: 208 3
   -- end of death anim, end of spawning: 80 4
   -- death sequence: playing, paused, endanim, spawning, playing
end

function getCheckpointDistance()
   -- runs from 0 to 255
   local low = 0xff - memory.readbyte(0x7a)
   -- runs from 0 to 11
   -- changes from 0 to 1 after taking a single step from spawn
   -- the boss area is at the end of section 11
   local high = 0xfd - memory.readbyte(0x7b)
   local pos = high + low / 256

   -- section 1 2 3 4 5 6 7 8 9 10 11
   -- spawns@ 1 3 3 5 5 5 5 7 9 11 11
   
   local offset, dist
   if pos < 2 then
      offset = 1 - pos -- relative position of spawnpoint
      dist = 2 - pos -- distance to next spawnpoint change
   elseif pos < 4 then
      offset = 3 - pos
      dist = 4 - pos
   elseif pos < 8 then
      offset = 5 - pos
      dist = 8 - pos
   elseif pos < 9 then
      offset = 7 - pos
      dist = 9 - pos
   elseif pos < 10 then
      offset = 9 - pos
      dist = 10 - pos
   else
      offset = 11 - pos
      dist = 0
   end
   return string.format('%2d %.2f %.2f', high, offset, dist)
end

function getPlayerPos()
   local x = memory.readbyte(0x62)
   local y = memory.readbyte(0x64)
   return {x=x, y=y}
end

function getPlayerAim()
   local x = memory.readbytesigned(0x76)
   local y = memory.readbytesigned(0x78)
   return {x=x, y=y}
end

main()
